import threading
import unittest
from http.server import BaseHTTPRequestHandler, HTTPServer
from unittest import mock, TestCase as BaseTestCase

import grpc
import sleuth
from google.api_core.client_options import ClientOptions
from google.api_core.exceptions import NotFound
from google.api_core.retry import Retry
from google.cloud.tasks_v2 import CloudTasksClient
from google.cloud.tasks_v2.gapic.transports.cloud_tasks_grpc_transport import \
    CloudTasksGrpcTransport

from server import _make_task_request, create_server


mock_calls = []


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self._handle_request()

    def do_POST(self):
        self._handle_request()

    def _handle_request(self):
        mock_calls.append({
            'method': self.command,
            'path': self.path,
            'headers': dict(self.headers),
        })
        self.send_response(200)
        self.end_headers()


class MockServer(threading.Thread):
    def __init__(self, port, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._port = port
        self.is_running = threading.Event()
        self._httpd = None

    def run(self):
        self._httpd = HTTPServer(("localhost", self._port), MockRequestHandler)
        self.is_running.set()
        self._httpd.serve_forever()

    def join(self, timeout=None):
        self.is_running.clear()
        if self._httpd:
            self._httpd.shutdown()
            self._httpd.server_close()


class TestCase(BaseTestCase):
    def setUp(self):
        self._server = create_server("localhost", 9022)
        self._server.start()

        transport = CloudTasksGrpcTransport(channel=grpc.insecure_channel("127.0.0.1:9022"))

        self._client = CloudTasksClient(
            transport=transport,
            client_options=ClientOptions(api_endpoint="127.0.0.1:9022")
        )

        self._parent = self._client.location_path('[PROJECT]', '[LOCATION]')

        # Create default queue
        self._client.create_queue(
            self._parent, {"name": "%s/queues/default" % self._parent},
            retry=Retry(initial=1e-6, maximum=5),  # Wait until server starts
        )

    def tearDown(self):
        self._server.stop()

    def test_create_queue(self):
        queue1_path = "%s/queues/test_queue1" % self._parent
        queue2_path = "%s/queues/test_queue2" % self._parent

        ret = self._client.create_queue(
            self._parent, {"name": queue1_path}
        )
        self.assertEqual(ret.name, queue1_path)

        ret = self._client.create_queue(
            self._parent, {"name": queue2_path}
        )
        self.assertEqual(ret.name, queue2_path)

        return (queue1_path, queue2_path)

    def test_list_queues(self):
        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
        self._client.delete_queue(path)

        self.test_create_queue()  # Create a couple of queues

        queues = self._client.list_queues(parent=self._parent)
        self.assertEqual(len(list(queues)), 2)

    def test_get_queue(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        queue = self._client.get_queue(path)
        self.assertEqual(queue.name, path)

    def test_delete_queue(self):
        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
        self._client.delete_queue(path)

        self.test_create_queue()  # Create a couple of queues

        queues = self._client.list_queues(parent=self._parent)
        self.assertEqual(len(list(queues)), 2)

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.delete_queue(path)

        queues = self._client.list_queues(parent=self._parent)
        self.assertEqual(len(list(queues)), 1)

    def test_pause_queue(self):
        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
        self._client.delete_queue(path)

        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        response = self._client.pause_queue(path)
        self.assertEqual(response.state, 2)

    def test_resume_queue(self):
        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
        self._client.delete_queue(path)

        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        response = self._client.pause_queue(path)
        self.assertEqual(response.state, 2)

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        response = self._client.resume_queue(path)
        self.assertEqual(response.state, 1)

    def test_purge_queue(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")

        # Pause the queue as we don't want tasks to be processed
        self._client.pause_queue(path)

        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'http_method': 'POST',
                'relative_uri': '/example_task_handler',
                'body': payload.encode()
            }
        }

        # Create 3 tasks
        self._client.create_task(path, task)
        self._client.create_task(path, task)
        self._client.create_task(path, task)

        tasks = [x for x in self._client.list_tasks(path)]
        self.assertEqual(len(tasks), 3)

        # Check again, make sure list_tasks didn't do anything
        tasks = [x for x in self._client.list_tasks(path)]
        self.assertEqual(len(tasks), 3)

        self._client.purge_queue(path)

        tasks = [x for x in self._client.list_tasks(path)]
        self.assertEqual(len(tasks), 0)

    def test_create_task(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")

        self._client.pause_queue(path)
        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'http_method': 'POST',
                'relative_uri': '/example_task_handler',
                'body': payload.encode()
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

    def test_create_task_without_http_method(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")

        self._client.pause_queue(path)
        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'relative_uri': '/example_task_handler',
                'body': payload.encode()
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

    def test_run_task(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path(
            '[PROJECT]', '[LOCATION]', "test_queue2"
        )

        self._client.pause_queue(path)  # Don't run any tasks while testing

        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # specify the type of request.
                'http_method': 'POST',
                'relative_uri': '/example_task_handler',
                'body': payload.encode()
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

        class FakeResponse:
            status = 200

        with sleuth.fake("server._make_task_request", return_value=FakeResponse()):
            self._client.run_task(response.name)

        # Should return NOT_FOUND
        self.assertRaises(
            NotFound,
            self._client.run_task,
            "%s/tasks/1119129292929292929" % path,  # Not a valid task
        )

    def test_default_queue_name(self):
        server = create_server("localhost", 9023, "localhost", 10124, ["projects/[P]/locations/[L]/queues/[Q]"])
        server.start()

        transport = CloudTasksGrpcTransport(channel=grpc.insecure_channel("127.0.0.1:9023"))
        client = CloudTasksClient(
            transport=transport,
            client_options=ClientOptions(api_endpoint="127.0.0.1:9023")
        )

        queues = list(client.list_queues(parent="projects/[P]/locations/[L]"))
        self.assertEqual(len(queues), 1)

        queue = queues[0]
        self.assertEqual(queue.name, "projects/[P]/locations/[L]/queues/[Q]")

        server.stop()

    def test_multiple_default_queue_names(self):
        server = create_server(
            "localhost", 9023, "localhost", 10124,
            ["projects/[P]/locations/[L]/queues/[Q]", "projects/[P]/locations/[L]/queues/[Q2]"]
        )
        server.start()

        transport = CloudTasksGrpcTransport(channel=grpc.insecure_channel("127.0.0.1:9023"))
        client = CloudTasksClient(
            transport=transport,
            client_options=ClientOptions(api_endpoint="127.0.0.1:9023")
        )

        queues = list(client.list_queues(parent="projects/[P]/locations/[L]"))
        self.assertEqual(len(queues), 2)

        self.assertEqual(queues[0].name, "projects/[P]/locations/[L]/queues/[Q]")
        self.assertEqual(queues[1].name, "projects/[P]/locations/[L]/queues/[Q2]")

        server.stop()

    def test_custom_headers(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")

        self._client.pause_queue(path)

        payload = "Hello World!"

        task = {
            'app_engine_http_request': {
                'http_method': 'POST',
                'relative_uri': '/example_task_handler',
                'body': payload.encode(),
                'headers': {'custom': 'custom'}
            }
        }

        self._client.create_task(path, task)
        parent = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        task_ = [x for x in self._client.list_tasks(parent)][-1]
        def noop(req): return req
        with mock.patch('urllib.request.urlopen', noop):
            req = _make_task_request('test_queue2', task_, "localhost", 9009)
        assert req.headers['Custom'] == 'custom'


class CustomPortTestCase(BaseTestCase):

    @classmethod
    def setUpClass(cls):
        cls._server = MockServer(10123)
        cls._server.start()
        cls._server.is_running.wait()

    @classmethod
    def tearDownClass(cls):
        cls._server.join(timeout=1)

    def setUp(self):
        mock_calls.clear()

        self._server = create_server("localhost", 9022, "localhost", 10123)
        self._server.start()

        transport = CloudTasksGrpcTransport(channel=grpc.insecure_channel("127.0.0.1:9022"))

        self._client = CloudTasksClient(
            transport=transport,
            client_options=ClientOptions(api_endpoint="127.0.0.1:9022")
        )

        self._parent = self._client.location_path('[PROJECT]', '[LOCATION]')

        # Create default queue
        self._client.create_queue(
            self._parent, {"name": "%s/queues/default" % self._parent},
            retry=Retry(initial=1e-6, maximum=5),  # Wait until server starts
        )

    def tearDown(self):
        self._server.stop()

    def test_create_queue(self):
        path1 = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue1")
        ret = self._client.create_queue(self._parent, {"name": path1})
        self.assertEqual(ret.name, path1)

        path2 = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        ret = self._client.create_queue(self._parent, {"name": path2})
        self.assertEqual(ret.name, path2)

    def test_custom_target_host(self):
        # Override the target host so we can check that works
        self._server._state._target_host = "www.example.com"

        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.pause_queue(path)  # Don't run any tasks while testing

        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'relative_uri': '/example_task_handler',
                'body': payload.encode(),
                'headers': {
                    'Content-Type': 'text/plain'
                },
            }
        }

        response = self._client.create_task(path, task)

        class FakeResponse:
            status = 200

        with sleuth.fake("server._make_task_request", return_value=FakeResponse()) as fake:
            self._client.run_task(response.name)
            self.assertEqual(fake.calls[0].args[2], "www.example.com")

    def test_run_app_engine_http_request_task(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.pause_queue(path)  # Don't run any tasks while testing

        payload = "Hello World!"

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'relative_uri': '/example_task_handler',
                'body': payload.encode(),
                'headers': {
                    'Content-Type': 'text/plain'
                },
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

        self._client.run_task(response.name)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0]['method'], 'POST')
        self.assertEqual(mock_calls[0]['path'], '/example_task_handler')
        self.assertEqual(mock_calls[0]['headers']['X-Appengine-Queuename'], path)
        self.assertEqual(mock_calls[0]['headers']['Content-Type'], 'text/plain')

    def test_run_app_engine_http_request_task_with_custom_method(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.pause_queue(path)  # Don't run any tasks while testing

        task = {
            'app_engine_http_request': {  # Specify the type of request.
                'http_method': 'GET',
                'relative_uri': '/example_task_handler',
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

        self._client.run_task(response.name)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0]['method'], 'GET')
        self.assertEqual(mock_calls[0]['path'], '/example_task_handler')
        self.assertEqual(mock_calls[0]['headers']['X-Appengine-Queuename'], path)

    def test_run_http_request_task(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.pause_queue(path)  # Don't run any tasks while testing

        payload = "Hello World!"

        task = {
            'http_request': {  # Specify the type of request.
                'url': 'http://localhost:10123/http_request_task_handler',
                'body': payload.encode(),
                'headers': {
                    'Content-Type': 'text/plain'
                },
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

        self._client.run_task(response.name)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0]['method'], 'POST')
        self.assertEqual(mock_calls[0]['path'], '/http_request_task_handler')
        self.assertEqual(mock_calls[0]['headers']['X-Cloudtasks-Queuename'], path)
        self.assertEqual(mock_calls[0]['headers']['Content-Type'], 'text/plain')

    def test_run_http_request_task_with_custom_method(self):
        self.test_create_queue()  # Create a couple of queues

        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
        self._client.pause_queue(path)  # Don't run any tasks while testing

        task = {
            'http_request': {  # Specify the type of request.
                'http_method': 'GET',
                'url': 'http://localhost:10123/http_request_task_handler',
            }
        }

        response = self._client.create_task(path, task)
        self.assertTrue(response.name.startswith(path))

        self._client.run_task(response.name)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0]['method'], 'GET')
        self.assertEqual(mock_calls[0]['path'], '/http_request_task_handler')
        self.assertEqual(mock_calls[0]['headers']['X-Cloudtasks-Queuename'], path)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import argparse
import logging
import sys

import yaml

from gcloud_tasks_emulator.server import create_server, DEFAULT_TARGET_PORT, DEFAULT_TARGET_HOST

# Random, apparently not often used
DEFAULT_PORT = 9022


def run_server(host, port, target_host, target_port, default_queue_names, max_retries):
    server = create_server(host, port, target_host, target_port, default_queue_names, max_retries)
    return server.run()


def read_queue_yaml(path):
    queues = []
    with open(path, "r") as f:
        try:
            data = yaml.safe_load(f)
            for queue in data['queue']:
                queues.append(queue['name'])

        except yaml.YAMLError:
            logging.exception("Error reading %s", path)
            return False, queues

    return True, queues


def prepare_args_parser():
    parser = argparse.ArgumentParser(description='Google Cloud Task Emulator')
    subparsers = parser.add_subparsers(title='subcommands', dest="subcommand")
    start = subparsers.add_parser('start', help='start the emulator')
    start.add_argument(
        "-p", "--port",
        type=int, help='the port to run the server on', default=DEFAULT_PORT
    )
    start.add_argument(
        "-t", "--target-port",
        type=int, help='the port to which the task runner will POST requests to',
        default=DEFAULT_TARGET_PORT,
    )
    start.add_argument(
        "-H", "--target-host",
        type=str, help="the hostname to submit tasks too (default is 'localhost')",
        default=DEFAULT_TARGET_HOST,
    )
    start.add_argument("-q", "--quiet", action="store_true", default=False)
    start.add_argument(
        "-d", "--default-queue", type=str, action="append",
        help="If specified will create a queue with the passed name. "
             "Name should be in the format of projects/PROJECT_ID/locations/LOCATION_ID/queues/QUEUE_ID"
    )
    start.add_argument(
        "-r", "--max-retries",
        type=int, help='maximum number of retries when a task is failed (default is infinity)',
        default=-1,
    )
    start.add_argument(
        "-Q", "--queue-yaml",
        type=str, help="path to a queue.yaml to initialize the default queues",
        default=None
    )
    start.add_argument(
        "-P", "--queue-yaml-project",
        type=str, help="project ID to use for queues created from queue-yaml",
        default="[PROJECT]"
    )
    start.add_argument(
        "-L", "--queue-yaml-location",
        type=str, help="location ID to use for queues created from queue-yaml",
        default="[LOCATION]"
    )
    return parser


def main():
    print("Starting Cloud Tasks Emulator")
    parser = prepare_args_parser()
    args = parser.parse_args()
    if args.subcommand != "start":
        parser.print_usage()
        sys.exit(1)
    root = logging.getLogger()
    root.addHandler(logging.StreamHandler())
    if args.quiet:
        root.setLevel(logging.ERROR)
    else:
        root.setLevel(logging.INFO)
    default_queues = set(args.default_queue or [])
    # FIXME: We simply read queue names from queue.yaml. Instead we should
    # read all queue parameters and handle them correctly in the emulator
    if args.queue_yaml:
        success, queues = read_queue_yaml(args.queue_yaml)
        queues = [
            "projects/%s/locations/%s/queues/%s" % (
                args.queue_yaml_project,
                args.queue_yaml_location,
                x
            ) for x in queues
        ]

        if success:
            default_queues = default_queues.union(set(queues))
        else:
            sys.exit(1)
    sys.exit(run_server("localhost", args.port, args.target_host, args.target_port, default_queues, args.max_retries))


if __name__ == '__main__':
    main()



import logging
import threading
import time
from concurrent import futures
from datetime import datetime
from typing import Dict
from urllib import error, request
from urllib.parse import parse_qs

import grpc
from google.api_core.exceptions import NotFound, FailedPrecondition, GoogleAPICallError
from google.cloud.tasks_v2.proto import (cloudtasks_pb2, cloudtasks_pb2_grpc,
                                         queue_pb2, task_pb2, target_pb2)
from google.protobuf import empty_pb2
from google.protobuf.timestamp_pb2 import Timestamp
from google.rpc.status_pb2 import Status

# Time to sleep between iterations of the threads
_LOOP_SLEEP_TIME = 0.1

DEFAULT_TARGET_PORT = 80
DEFAULT_TARGET_HOST = "localhost"

Queue = queue_pb2.Queue
Task = task_pb2.Task
Attempt = task_pb2.Attempt

logger = logging.getLogger("gcloud-tasks-emulator")


def _make_task_request(queue_name: str, task: Task, host: str, port: int):
    logger.info("[TASKS] Submitting task %s", task.name)

    headers = {}
    data = None

    if task.app_engine_http_request.relative_uri:
        method = target_pb2.HttpMethod.Name(task.app_engine_http_request.http_method)
        data = task.app_engine_http_request.body

        url = "http://%s:%d%s" % (
            host,
            port,
            task.app_engine_http_request.relative_uri
        )

        headers.update(getattr(task.app_engine_http_request, 'headers', {}))
        headers.update({
            'X-AppEngine-QueueName': queue_name,
            'X-AppEngine-TaskName': task.name.rsplit("/", 1)[-1],
            'X-AppEngine-TaskRetryCount': task.dispatch_count - 1,
            'X-AppEngine-TaskExecutionCount': task.dispatch_count - 1,
            'X-AppEngine-TaskETA': task.schedule_time.ToSeconds()
        })
    elif task.http_request.url:
        method = target_pb2.HttpMethod.Name(task.http_request.http_method)
        data = task.http_request.body
        url = task.http_request.url
        headers.update(getattr(task.http_request, "headers", {}))
        headers.update({
            'X-CloudTasks-QueueName': queue_name,
            'X-CloudTasks-TaskName': task.name.rsplit("/", 1)[-1],
            'X-CloudTasks-TaskRetryCount': task.dispatch_count - 1,
            'X-CloudTasks-TaskExecutionCount': task.dispatch_count - 1,
            'X-CloudTasks-TaskETA': task.schedule_time.ToSeconds()
        })
    else:
        raise Exception("Either app_engine_http_request or http_request is required")

    req = request.Request(url, method=method, data=data)
    for k, v in headers.items():
        req.add_header(k, v)

    logger.info("[TASKS] Requesting %s %s", req.method, req.full_url)
    return request.urlopen(req)


class QueueState(object):
    """
        Keeps the state of queues and tasks in memory
        so they can be processed
    """

    def __init__(self, target_host: str, target_port: int, max_retries: int):
        self._queues: Dict[str, Queue] = {}
        self._queue_tasks: Dict[str, Task] = {}
        self._target_host: str = target_host
        self._target_port: int = target_port
        self._max_retries: int = max_retries

    def create_queue(self, parent, queue: Queue):
        assert queue.name
        if queue.name not in self._queues:
            self._queues[queue.name] = queue
            self._queues[queue.name].state = (
                queue_pb2._QUEUE_STATE.values_by_name["RUNNING"].number
            )
            self._queue_tasks[queue.name] = []
            logger.info("[TASKS] Created queue %s", queue.name)

        return self._queues[queue.name]

    def create_task(self, queue: Queue, task: Task):
        task.name = task.name or "%s/tasks/%s" % (
            queue, int(datetime.now().timestamp() * 1000000)
        )

        if task.app_engine_http_request.relative_uri:
            # Set a default http_method
            task.app_engine_http_request.http_method = (
                task.app_engine_http_request.http_method or target_pb2.HttpMethod.Value("POST")
            )
        elif task.http_request.url:
            # Set a default http_method
            task.http_request.http_method = (
                task.http_request.http_method or target_pb2.HttpMethod.Value("POST")
            )

        if queue not in self._queue_tasks:
            raise FailedPrecondition(f"Queue {queue} does not exist.")

        self._queue_tasks[queue].append(task)
        logger.info("[TASKS] Created task %s", task.name)
        return task

    def purge_queue(self, queue_name: str):
        if queue_name in self._queues:
            # Wipe the tasks out
            logger.info("[TASKS] Purging queue %s", queue_name)
            self._queue_tasks[queue_name] = []
            return self._queues[queue_name]
        else:
            logger.error(
                "[TASKS] Tried to purge an invalid queue: %s",
                queue_name
            )
            raise ValueError("Invalid queue: %s" % queue_name)

    def pause_queue(self, queue_name: str):
        if queue_name in self._queues:
            logger.info("[TASKS] Pausing queue %s", queue_name)
            self._queues[queue_name].state = (
                queue_pb2._QUEUE_STATE.values_by_name["PAUSED"].number
            )
            return self._queues[queue_name]
        else:
            logger.error(
                "[TASKS] Tried to pause an invalid queue: %s",
                queue_name
            )
            raise ValueError("Invalid queue: %s" % queue_name)

    def resume_queue(self, queue_name: str):
        if queue_name in self._queues:
            logger.info("[TASKS] Resuming queue %s", queue_name)
            self._queues[queue_name].state = (
                queue_pb2._QUEUE_STATE.values_by_name["RUNNING"].number
            )
            return self._queues[queue_name]
        else:
            logger.error(
                "[TASKS] Tried to resume an invalid queue: %s",
                queue_name
            )
            raise ValueError("Invalid queue: %s" % queue_name)

    def list_tasks(self, queue_name: str):
        return self._queue_tasks[queue_name]

    def queue_names(self):
        return list(self._queues)

    def queues(self):
        return list(self._queues.values())

    def queue(self, name: str) -> Queue:
        return self._queues[name]

    def delete_queue(self, name: str):
        if name in self._queues:
            logger.info("[TASKS] Deleting queue %s", name)
            del self._queues[name]

        if name in self._queue_tasks:
            del self._queue_tasks[name]

    def submit_task(self, task_name: str, force_run: bool = False) -> Task:
        """
            Actually executes a task. If force_run is True then the scheduled
            time will be ignored. This is used mainly for the RunTask API call.
        """

        try:
            queue_name = task_name.rsplit("/", 2)[0]
            if queue_name not in self._queue_tasks:
                raise ValueError("Not a valid queue")
        except IndexError:
            # Invalid task name, raise ValueError
            raise ValueError()

        # This is a special-case that does not exist
        # one the live server and it exists so that
        # local development servers can direct a task
        # to run on a particular port
        qs = task_name.rsplit("?", 1)[-1]
        if qs:
            params = parse_qs(qs)
            port = int(params.get("port", [self._target_port])[0])
            task_name = task_name.rsplit("?", 1)[0]
        else:
            port = self._target_port

        index = None

        # Locate the task in the queue
        for i, task in enumerate(self._queue_tasks[queue_name]):
            if task.name == task_name:
                index = i
                break
        else:
            logger.debug(
                "[TASKS] Tasks were: %s",
                [x.name for x in self._queue_tasks[queue_name]]
            )
            raise NotFound("Task not found: %s" % task_name)

        def now():
            current_time = Timestamp()
            current_time.GetCurrentTime()
            return current_time

        schedule_time = now()
        dispatch_time = None
        response_time = None
        response_status = 200

        task: Task = self._queue_tasks[queue_name].pop(index)  # Remove the task

        if (
            (not force_run) and
            task.HasField("schedule_time") and
            task.schedule_time.ToDatetime() >= schedule_time.ToDatetime()
        ):
            logger.info(
                "[TASKS] Task %s is scheduled for future execution. Moving it to the end of the queue.", task_name
            )
            self._queue_tasks[queue_name].append(task)
            return task

        task.dispatch_count += 1
        try:
            dispatch_time = now()
            response = _make_task_request(queue_name, task, self._target_host, port)
        except error.HTTPError as e:
            response_status = e.code
            logger.error("[TASKS] Error submitting task, reason: %s", e.reason)
        except (ConnectionRefusedError, error.URLError):
            response_status = 500
            logger.error("[TASKS] Error submitting task %s", task_name)
        else:
            response_status = response.status

        attempt = Attempt(
            schedule_time=schedule_time,
            dispatch_time=dispatch_time,
            response_time=response_time,
            response_status=Status(code=response_status)
        )

        kwargs = {
            "first_attempt": task.first_attempt or attempt,
            "last_attempt": attempt
        }

        task.MergeFrom(Task(**kwargs))

        assert task

        if 400 <= response_status < 600:
            if self._max_retries < 0 or task.dispatch_count <= self._max_retries:
                logger.info("[TASKS] Moving failed task %s to the back of the queue.", task_name)
                self._queue_tasks[queue_name].append(task)
            else:
                logger.info("[TASKS] Giving up failed task %s", task_name)

        return task


class Greeter(cloudtasks_pb2_grpc.CloudTasksServicer):
    def __init__(self, state: QueueState):
        super().__init__()

        self._state: QueueState = state

    def CreateQueue(self, request, context):
        try:
            return self._state.create_queue(request.parent, request.queue)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def ListQueues(self, request, context):
        try:
            queues = [x for x in self._state.queues() if x.name.startswith(request.parent)]
            return cloudtasks_pb2.ListQueuesResponse(queues=queues)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def GetQueue(self, request, context):
        try:
            return self._state.queue(request.name)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def PauseQueue(self, request, context):
        try:
            return self._state.pause_queue(request.name)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def ResumeQueue(self, request, context):
        try:
            return self._state.resume_queue(request.name)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def PurgeQueue(self, request, context):
        try:
            return self._state.purge_queue(request.name)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def ListTasks(self, request, context):
        try:
            return cloudtasks_pb2.ListTasksResponse(
                tasks=self._state.list_tasks(request.parent)
            )
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def DeleteQueue(self, request, context):
        try:
            self._state.delete_queue(request.name)
            return empty_pb2.Empty()
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def CreateTask(self, request, context):
        try:
            return self._state.create_task(request.parent, request.task)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)

    def RunTask(self, request, context):
        try:
            return self._state.submit_task(request.name, force_run=True)
        except GoogleAPICallError as err:
            context.abort(err.grpc_status_code, err.message)


class APIThread(threading.Thread):
    def __init__(self, state: QueueState, host: str, port: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._state = state
        self._port = port
        self._is_running = threading.Event()
        self._httpd = None

    def run(self):
        self._httpd = grpc.server(futures.ThreadPoolExecutor(max_workers=2))
        cloudtasks_pb2_grpc.add_CloudTasksServicer_to_server(
            Greeter(self._state), self._httpd
        )

        interface = '[::]:%s' % self._port
        self._httpd.add_insecure_port(interface)

        logger.info("[API] Starting API server at %s", interface)
        self._httpd.start()

        while self._is_running.is_set():
            time.sleep(_LOOP_SLEEP_TIME)

    def join(self, timeout=None):
        self._is_running.clear()
        if self._httpd:
            self._httpd.stop(grace=0)
        logger.info("[API] Stopping API server")


class Processor(threading.Thread):
    def __init__(self, state: QueueState):
        super().__init__()

        self._state: QueueState = state
        self._is_running = threading.Event()
        self._known_queues = set()
        self._queue_threads = {}

    def run(self):
        self._is_running.set()

        logger.info("[PROCESSOR] Starting task processor")
        while self._is_running.is_set():
            queue_names = self._state.queue_names()
            for queue in queue_names:
                self.process_queue(queue)
            time.sleep(_LOOP_SLEEP_TIME)

    def _process_queue(self, queue: str):
        while self._is_running.is_set():
            # Queue was deleted, stop processing
            if queue not in self._state._queues:
                break

            if queue not in self._state._queue_tasks:
                break

            if self._state.queue(queue).state == queue_pb2._QUEUE_STATE.values_by_name["RUNNING"].number:
                tasks = self._state._queue_tasks[queue][:]
                while tasks:
                    task = tasks.pop(0)
                    logger.info("[TASKS] Processing next task %s", task.name)
                    self._state.submit_task(task.name)

            time.sleep(_LOOP_SLEEP_TIME)

    def process_queue(self, queue_name: str):
        if queue_name not in self._known_queues:
            # A queue was just created
            self._known_queues.add(queue_name)

            thread = threading.Thread(
                target=self._process_queue, args=[queue_name]
            )
            self._queue_threads[queue_name] = thread
            self._queue_threads[queue_name].start()

    def join(self, timeout=None):
        self._is_running.clear()
        for thread in self._queue_threads.values():
            if thread.is_alive():
                thread.join(timeout=0)

        super().join(timeout)


class Server(object):
    def __init__(self, host, port, target_host, target_port, default_queue_names, max_retries):
        self._state = QueueState(target_host, target_port, max_retries)
        self._api = APIThread(self._state, host, port)
        self._processor = Processor(self._state)

        for default_queue_name in default_queue_names:
            parent = default_queue_name.rsplit("/", 3)[0]
            self._state.create_queue(
                parent, Queue(name=default_queue_name)
            )

    def start(self):
        self._api.start()  # Start the API thread
        self._processor.start()

    def stop(self):
        self._processor.join(timeout=1)
        self._api.join(timeout=1)

    def run(self):
        try:
            self.start()

            logger.info("[SERVER] All services started")

            while True:
                try:
                    time.sleep(_LOOP_SLEEP_TIME)
                except KeyboardInterrupt:
                    break

        finally:
            self.stop()


def create_server(
    host, port,
    target_host=DEFAULT_TARGET_HOST, target_port=DEFAULT_TARGET_PORT,
    default_queue_names=None, max_retries=-1
):
    return Server(host, port, target_host, target_port, default_queue_names or [], max_retries)
